
/******************************************************************************
* MODULE     : dyadic.hpp
* DESCRIPTION: Dyadic numbers
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_DYADIC_HPP
#define __MML_DYADIC_HPP
#include "mml_gmp.hpp"
namespace mml {

class dyadic_rep {
private:
  mpfr_t x;

public:
  inline dyadic_rep () { mpfr_init (x); }
  inline dyadic_rep (mp_prec_t prec) { mpfr_init2 (x, prec); }
  inline dyadic_rep (mpfr_t& x1) {
    mpfr_init2 (x, mpfr_get_prec (x1)); }
  inline dyadic_rep (mpfr_t& x1, mpfr_t& x2) {
    mpfr_init2 (x, min (mpfr_get_prec (x1), mpfr_get_prec (x2))); }
  inline ~dyadic_rep () { mpfr_clear (x); }

  friend class dyadic;
  friend class interval;
};

class dyadic_rep;
class dyadic {
INDIRECT_PROTO (dyadic, dyadic_rep)
private:
  inline dyadic (mpfr_t& x): rep (new dyadic_rep (x)) {}
  inline dyadic (mpfr_t& x, mpfr_t& y): rep (new dyadic_rep (x, y)) {}
  inline mpfr_t& operator * () const { return rep->x; }

public:
  // constructors for default precision
  inline dyadic (signed int i):
    rep (new dyadic_rep ()) { mpfr_set_si (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (unsigned int i):
    rep (new dyadic_rep ()) { mpfr_set_ui (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (signed short int i):
    rep (new dyadic_rep ()) { mpfr_set_si (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (unsigned short int i):
    rep (new dyadic_rep ()) { mpfr_set_ui (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (signed long int i):
    rep (new dyadic_rep ()) { mpfr_set_si (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (unsigned long int i):
    rep (new dyadic_rep ()) { mpfr_set_ui (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (float x2):
    rep (new dyadic_rep ()) { mpfr_set_d (rep->x, x2, DEFAULT_ROUND); }
  inline dyadic (double x2):
    rep (new dyadic_rep ()) { mpfr_set_d (rep->x, x2, DEFAULT_ROUND); }

  // constructors for user provided precision
  inline dyadic (const dyadic& x, mp_prec_t prec):
    rep (new dyadic_rep (prec)) { mpfr_set (rep->x, x->x, DEFAULT_ROUND); }
  inline dyadic (signed int i, mp_prec_t prec):
    rep (new dyadic_rep (prec)) { mpfr_set_si (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (unsigned int i, mp_prec_t prec):
    rep (new dyadic_rep (prec)) { mpfr_set_ui (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (signed short int i, mp_prec_t prec):
    rep (new dyadic_rep (prec)) { mpfr_set_si (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (unsigned short int i, mp_prec_t prec):
    rep (new dyadic_rep (prec)) { mpfr_set_ui (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (signed long int i, mp_prec_t prec):
    rep (new dyadic_rep (prec)) { mpfr_set_si (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (unsigned long int i, mp_prec_t prec):
    rep (new dyadic_rep (prec)) { mpfr_set_ui (rep->x, i, DEFAULT_ROUND); }
  inline dyadic (float x2, mp_prec_t prec):
    rep (new dyadic_rep (prec)) { mpfr_set_d (rep->x, x2, DEFAULT_ROUND); }
  inline dyadic (double x2, mp_prec_t prec):
    rep (new dyadic_rep (prec)) { mpfr_set_d (rep->x, x2, DEFAULT_ROUND); }
  inline dyadic (mp_limb_t i, mp_prec_t prec, mp_exp_t exp):
    rep (new dyadic_rep (prec)) {
    mpfr_set_ui (rep->x, i, DEFAULT_ROUND);
    mpfr_mul_2si (rep->x, rep->x, exp, DEFAULT_ROUND); }

  // assignment
  inline dyadic& operator = (signed int i) {
    mpfr_set_si (rep->x, i, DEFAULT_ROUND); return *this; }
  inline dyadic& operator = (unsigned int i) {
    mpfr_set_ui (rep->x, i, DEFAULT_ROUND); return *this; }
  inline dyadic& operator = (signed short int i) {
    mpfr_set_si (rep->x, i, DEFAULT_ROUND); return *this; }
  inline dyadic& operator = (unsigned short int i) {
    mpfr_set_ui (rep->x, i, DEFAULT_ROUND); return *this; }
  inline dyadic& operator = (signed long int i) {
    mpfr_set_si (rep->x, i, DEFAULT_ROUND); return *this; }
  inline dyadic& operator = (unsigned long int i) {
    mpfr_set_ui (rep->x, i, DEFAULT_ROUND); return *this; }
  inline dyadic& operator = (float x2) {
    mpfr_set_d (rep->x, x2, DEFAULT_ROUND); return *this; }
  inline dyadic& operator = (double x2) {
    mpfr_set_d (rep->x, x2, DEFAULT_ROUND); return *this; }

  // input/output
  inline friend ostream& operator << (ostream& out, const dyadic& x) {
    return out << (*x); }

  // basic arithmetic
  inline friend dyadic operator - (const dyadic& x1) {
    dyadic r (*x1); mpfr_neg (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic operator + (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_add (*r, *x1, *x2, DEFAULT_ROUND); return r; }
  inline friend dyadic operator - (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_sub (*r, *x1, *x2, DEFAULT_ROUND); return r; }
  inline friend dyadic operator * (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_mul (*r, *x1, *x2, DEFAULT_ROUND); return r; }
  inline friend dyadic operator / (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_div (*r, *x1, *x2, DEFAULT_ROUND); return r; }
  inline friend dyadic operator << (const dyadic& x1, mp_exp_t shift) {
    dyadic r (*x1); mpfr_mul_2si (*r, *x1, shift, DEFAULT_ROUND); return r; }
  inline friend dyadic operator >> (const dyadic& x1, mp_exp_t shift) {
    dyadic r (*x1); mpfr_mul_2si (*r, *x1, -shift, DEFAULT_ROUND); return r; }
  inline friend dyadic& operator += (dyadic& x1, const dyadic& x2) {
    mpfr_add (*x1, *x1, *x2, DEFAULT_ROUND); return x1; }
  inline friend dyadic& operator -= (dyadic& x1, const dyadic& x2) {
    mpfr_sub (*x1, *x1, *x2, DEFAULT_ROUND); return x1; }
  inline friend dyadic& operator *= (dyadic& x1, const dyadic& x2) {
    mpfr_mul (*x1, *x1, *x2, DEFAULT_ROUND); return x1; }
  inline friend dyadic& operator /= (dyadic& x1, const dyadic& x2) {
    mpfr_div (*x1, *x1, *x2, DEFAULT_ROUND); return x1; }
  inline friend dyadic& operator <<= (dyadic& x1, mp_exp_t shift) {
    mpfr_mul_2si (*x1, *x1, shift, DEFAULT_ROUND); return x1; }
  inline friend dyadic& operator >>= (dyadic& x1, mp_exp_t shift) {
    mpfr_mul_2si (*x1, *x1, -shift, DEFAULT_ROUND); return x1; }
  inline friend dyadic sqrt (const dyadic& x1) {
    dyadic r (*x1); mpfr_sqrt (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic cbrt (const dyadic& x1) {
    dyadic r (*x1); mpfr_cbrt (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic pow (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_div (*r, *x1, *x2, DEFAULT_ROUND); return r; }
  inline friend dyadic abs (const dyadic& x1) {
    dyadic r (*x1); mpfr_abs (*r, *x1, DEFAULT_ROUND); return r; }

  // comparing dyadic numbers
  inline friend bool operator == (const dyadic& x1, const dyadic& x2) {
    return mpfr_cmp (*x1, *x2) == 0; }
  inline friend bool operator != (const dyadic& x1, const dyadic& x2) {
    return mpfr_cmp (*x1, *x2) != 0; }
  inline friend bool operator < (const dyadic& x1, const dyadic& x2) {
    return mpfr_cmp (*x1, *x2) < 0; }
  inline friend bool operator > (const dyadic& x1, const dyadic& x2) {
    return mpfr_cmp (*x1, *x2) > 0; }
  inline friend bool operator <= (const dyadic& x1, const dyadic& x2) {
    return mpfr_cmp (*x1, *x2) <= 0; }
  inline friend bool operator >= (const dyadic& x1, const dyadic& x2) {
    return mpfr_cmp (*x1, *x2) >= 0; }
  inline friend int sign (const dyadic& x1) {
    int r= mpfr_sgn (*x1); return (r<0? -1: (r==0? 0: 1)); }
  inline friend dyadic min (const dyadic& x1, const dyadic& x2) {
    return x1 <= x2? x1: x2; }
  inline friend dyadic max (const dyadic& x1, const dyadic& x2) {
    return x1 >= x2? x1: x2; }

  // special functions
  inline friend dyadic log (const dyadic& x1) {
    dyadic r (*x1); mpfr_log (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic log2 (const dyadic& x1) {
    dyadic r (*x1); mpfr_log2 (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic log10 (const dyadic& x1) {
    dyadic r (*x1); mpfr_log10 (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic exp (const dyadic& x1) {
    dyadic r (*x1); mpfr_exp (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic exp2 (const dyadic& x1) {
    dyadic r (*x1); mpfr_exp2 (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic cos (const dyadic& x1) {
    dyadic r (*x1); mpfr_cos (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic sin (const dyadic& x1) {
    dyadic r (*x1); mpfr_sin (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic tan (const dyadic& x1) {
    dyadic r (*x1); mpfr_tan (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic acos (const dyadic& x1) {
    dyadic r (*x1); mpfr_acos (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic asin (const dyadic& x1) {
    dyadic r (*x1); mpfr_asin (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic atan (const dyadic& x1) {
    dyadic r (*x1); mpfr_atan (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic cosh (const dyadic& x1) {
    dyadic r (*x1); mpfr_cosh (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic sinh (const dyadic& x1) {
    dyadic r (*x1); mpfr_sinh (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic tanh (const dyadic& x1) {
    dyadic r (*x1); mpfr_tanh (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic acosh (const dyadic& x1) {
    dyadic r (*x1); mpfr_acosh (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic asinh (const dyadic& x1) {
    dyadic r (*x1); mpfr_asinh (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic atanh (const dyadic& x1) {
    dyadic r (*x1); mpfr_atanh (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic gamma (const dyadic& x1) {
    dyadic r (*x1); mpfr_gamma (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic zeta (const dyadic& x1) {
    dyadic r (*x1); mpfr_zeta (*r, *x1, DEFAULT_ROUND); return r; }
  inline friend dyadic erf (const dyadic& x1) {
    dyadic r (*x1); mpfr_erf (*r, *x1, DEFAULT_ROUND); return r; }

  // constants
  inline static dyadic infinity () {
    dyadic r= new dyadic_rep ();
    mpfr_set_inf (*r, 1); return r; }
  inline static dyadic minus_infinity () {
    dyadic r= new dyadic_rep ();
    mpfr_set_inf (*r, -1); return r; }
  inline static dyadic log2 () {
    dyadic r= new dyadic_rep ();
    mpfr_const_log2 (*r, DEFAULT_ROUND); return r; }
  inline static dyadic pi () {
    dyadic r= new dyadic_rep ();
    mpfr_const_pi (*r, DEFAULT_ROUND); return r; }
  inline static dyadic euler () {
    dyadic r= new dyadic_rep ();
    mpfr_const_euler (*r, DEFAULT_ROUND); return r; }
  inline static dyadic infinity (mp_prec_t prec) {
    dyadic r= new dyadic_rep (prec);
    mpfr_set_inf (*r, 1); return r; }
  inline static dyadic minus_infinity (mp_prec_t prec) {
    dyadic r= new dyadic_rep (prec);
    mpfr_set_inf (*r, -1); return r; }
  inline static dyadic log2 (mp_prec_t prec) {
    dyadic r= new dyadic_rep (prec);
    mpfr_const_log2 (*r, DEFAULT_ROUND); return r; }
  inline static dyadic pi (mp_prec_t prec) {
    dyadic r= new dyadic_rep (prec);
    mpfr_const_pi (*r, DEFAULT_ROUND); return r; }
  inline static dyadic euler (mp_prec_t prec) {
    dyadic r= new dyadic_rep (prec);
    mpfr_const_euler (*r, DEFAULT_ROUND); return r; }

  // rounded arithmetic
  inline friend dyadic add_lo (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_add (*r, *x1, *x2, GMP_RNDD); return r; }
  inline friend dyadic add_up (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_add (*r, *x1, *x2, GMP_RNDU); return r; }
  inline friend dyadic sub_lo (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_sub (*r, *x1, *x2, GMP_RNDD); return r; }
  inline friend dyadic sub_up (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_sub (*r, *x1, *x2, GMP_RNDU); return r; }
  inline friend dyadic mul_lo (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_mul (*r, *x1, *x2, GMP_RNDD); return r; }
  inline friend dyadic mul_up (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_mul (*r, *x1, *x2, GMP_RNDU); return r; }
  inline friend dyadic div_lo (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_div (*r, *x1, *x2, GMP_RNDD); return r; }
  inline friend dyadic div_up (const dyadic& x1, const dyadic& x2) {
    dyadic r (*x1, *x2); mpfr_div (*r, *x1, *x2, GMP_RNDU); return r; }

  // miscellaneous functions
  inline friend bool is_zero (const dyadic& x) {
    return mpfr_sgn (*x) == 0; }
  inline friend bool is_infinity (const dyadic& x) {
    return mpfr_inf_p (*x) != 0; }
  inline friend bool is_nan (const dyadic& x) {
    return mpfr_nan_p (*x) != 0; }
  inline friend bool has_low_prec (const dyadic& x) {
    return mpfr_get_prec (*x) <= DEFAULT_LOW; }
  inline friend bool has_high_prec (const dyadic& x) {
    return mpfr_get_prec (*x) > DEFAULT_LOW; }
  inline friend mp_exp_t expo (const dyadic& x) {
    return mpfr_get_exp (*x); }
  inline friend mp_exp_t prec (const dyadic& x) {
    return (mp_exp_t) mpfr_get_prec (*x); }
  inline friend mp_exp_t step (const dyadic& x) {
    return mpfr_get_exp (*x) - mpfr_get_prec (*x); }

  // for interface with intervals
  friend class interval;
  friend class interval_rep;
  friend interval segment (const dyadic& min, const dyadic& max);
  friend interval ball (const dyadic& mid, const dyadic& rad);
  friend dyadic center (const interval& x);
  friend dyadic radius (const interval& x);
  friend dyadic lower (const interval& x);
  friend dyadic upper (const interval& x);
  friend dyadic bound (const interval& x);
  friend bool wider (const interval& x, const dyadic& rad);
  friend ostream& operator << (ostream& out, const interval& x);
  inline friend dyadic error (mp_limb_t k, mp_exp_t p= -DEFAULT_LOW) {
    dyadic r= new dyadic_rep (DEFAULT_LOW);
    mpfr_set_ui_2si (*r, k, p, GMP_RNDN); return r; }

  // for interface with reals
  static dyadic rndtol; // 1 - 16*2^DEFAULT_LOW (for rounding errors)
  inline friend dyadic spread (const dyadic& err) {
    dyadic r (*err);
    mpfr_mul (*r, *err, *rndtol, GMP_RNDD);
    return r; }
  inline friend dyadic spread (const dyadic& err,
			       const dyadic& part, const dyadic& tot) {
    dyadic r (*err);
    mpfr_mul (*r, *err, *rndtol, GMP_RNDD);
    mpfr_mul (*r, *err, *part, GMP_RNDD);
    mpfr_div (*r, *r, *tot, GMP_RNDD);
    return r; }
};
INDIRECT_IMPL (dyadic, dyadic_rep)

extern dyadic lp_epsilon;
extern dyadic lp_infinity;

}
#endif // __MML_DYADIC_HPP
